#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <CL/cl.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <algorithm>
#include <cstdlib>
#include <vector>
#include <windows.h>
//#include "Timer.h"
//#include <sys/time.h>
//#include <unistd.h>

// preprocessor defines
typedef unsigned int type;
typedef std::vector<type> vector;

//data size
#define DATA_SIZE 1<<25
#define NLOOPS 5000

//cpu input vectors
vector input(DATA_SIZE);
vector result_cpu(DATA_SIZE);

LARGE_INTEGER frequency;
LARGE_INTEGER start;
LARGE_INTEGER stop;



char* myLoadProgSource(const char* cFilename, const char* cPreamble, size_t* szFinalLength)
{
    // locals 
    FILE* pFileStream = NULL;
    size_t szSourceLength;

    pFileStream = fopen(cFilename, "rb");
    if(pFileStream == 0) 
    {       
        return NULL;
    }

    size_t szPreambleLength = strlen(cPreamble);

    // get the length of the source code
    fseek(pFileStream, 0, SEEK_END); 
    szSourceLength = ftell(pFileStream);
    fseek(pFileStream, 0, SEEK_SET); 

    // allocate a buffer for the source code string and read it in
    char* cSourceString = (char *)malloc(szSourceLength + szPreambleLength + 1); 
    memcpy(cSourceString, cPreamble, szPreambleLength);
    if (fread((cSourceString) + szPreambleLength, szSourceLength, 1, pFileStream) != 1)
    {
        fclose(pFileStream);
        free(cSourceString);
        return 0;
    }

    // close the file and return the total length of the combined (preamble + source) string
    fclose(pFileStream);
    if(szFinalLength != 0)
    {
        *szFinalLength = szSourceLength + szPreambleLength;
    }
    cSourceString[szSourceLength + szPreambleLength] = '\0';

    return cSourceString;
}

type filter(type value)
{
	return value >> 8;
}

void project_cpu(size_t elements)
{
	QueryPerformanceFrequency(&frequency);
	int nloops = NLOOPS;
	QueryPerformanceCounter(&start);
	//while(nloops--){
		
		for(size_t i = 0; i < elements; ++i)
		{
			type value = input[i];		
			result_cpu[i] = filter(value);
		}
		
	//}
	QueryPerformanceCounter(&stop);
	double timespent = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
	printf ("project on CPU took total = %f seconds\n", timespent);

	/*for(size_t i = 0; i != elements; ++i)
	{
		printf("input_cpu %u ",input[i]);
		printf("output_cpu %u ",result_cpu[i]);
	}
	printf("\n");*/

	//double sec = 
	//sec = sec/NLOOPS;
	//printf ("CPU projection took total time = %f seconds\n", sec);

	size_t elementsAccessed = 2 * elements;
	float GBs = (((elementsAccessed) * sizeof(type)) / (timespent)) / ((float)(1<<30));
	std::cout << " " << GBs << " GBs\n";
}





int main(int argc, char** const argv)
{

	size_t elements = DATA_SIZE;
	size_t d = 1;
	//std::cin >> elements;
	//OpenCL variables
	cl_context context;
	cl_context_properties properties[3];
	cl_kernel kernel;
	cl_command_queue command_queue;
	cl_program program;
	cl_int err;
	cl_uint num_of_platforms;
	cl_platform_id platform_id = NULL;
	//cl_device_id device_id[2];
	cl_uint num_of_devices = 0;
	
	cl_mem clInput;	
	cl_mem clOutput;
	
	cl_event kernel_event;
	cl_event map_event;
	cl_event unmap_event; 

	size_t global;
	size_t local;
	
	float nwork_groups;
	size_t work_group_size;
	size_t work_item_size[3];

	


	for(d=0; d<2; d++){
		for(elements=1024; elements<=33554432; elements*=2){

			printf("elements: %lu ",elements);
			printf("device: %lu ",d);

			input.resize(elements);
			result_cpu.resize(elements);
			//GPU input and output arrays
			type* inputData = new type[elements];
			type* results = new type[elements];

			srand((unsigned)time(0));
	
			//populating input with random values
			for(int sz=0; sz<elements; sz++)
			{
				inputData[sz] = rand() % RAND_MAX;
				input[sz] = inputData[sz];
				//clear the result array
				results[sz]=0;
				////printf("%u ",inputData[sz]);
			}
			////printf("\n");
	

			//calling cpu version of project
			project_cpu(elements);

			//(1) retreive the platform
			cl_uint numPlatforms;	//the NO. of platforms
			//cl_platform_id platform = NULL;	//the chosen platform
			cl_int	status = clGetPlatformIDs(0, NULL, &numPlatforms);
			if (status != CL_SUCCESS)
			{
				printf("Error: Getting platforms!");
				return 1;
			}

			/* choose the first available platform. */
			if(numPlatforms > 0)
			{
				cl_platform_id* platforms = (cl_platform_id* )malloc(numPlatforms* sizeof(cl_platform_id));
				status = clGetPlatformIDs(numPlatforms, platforms, NULL);
				platform_id = platforms[0];
				free(platforms);
			}

			//get platform name
			char tmpBuffer[1024];
			if ((err = clGetPlatformInfo(platform_id, CL_PLATFORM_NAME, sizeof(tmpBuffer), tmpBuffer, NULL)) != CL_SUCCESS)
			{
				printf("%d, Unable to get platform name!\n", err);
				return 1;
			}
			//printf("Platform name: %s\n", tmpBuffer);

			//get platform version
			if (clGetPlatformInfo(platform_id, CL_PLATFORM_VERSION, sizeof(tmpBuffer), tmpBuffer, NULL) != CL_SUCCESS)
			{
				printf("Unable to get platform version!\n");
				return 1;
			}
			//printf("Platform version: %s\n", tmpBuffer);
 
			//(2) try to get GPU device(s)

			// Get number of devices available 
			//cl_uint deviceount = 0;
			if(err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ALL, 0, NULL, &num_of_devices) != CL_SUCCESS)
			{
				printf("Unable to get device_id, %d\n", err);
				return 1;
			}

			cl_device_id* device_id = (cl_device_id*)malloc(sizeof(cl_device_id) * num_of_devices);
	
			if(num_of_devices == 0)
			{
				printf("No devices found!\n");
				return 1;
			}
			//printf("Number of devices: %d\n", num_of_devices);

			if ((err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_ALL, num_of_devices, device_id, NULL)) != CL_SUCCESS)
			{
				printf("clGetDeviceIDs failed, %d\n", err);
				return 1;
			}
	
			//get device name

			if (clGetDeviceInfo(device_id[d], CL_DEVICE_NAME, sizeof(tmpBuffer), &tmpBuffer, NULL) != CL_SUCCESS)
			{
				printf("Unable to get device name\n");
				return 1;
			} 
			printf("Device %u Name: %s\n", 0, tmpBuffer);


			//get max compute units in device
			cl_uint comp_unit;   
			err = clGetDeviceInfo(device_id[d], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), &comp_unit, NULL);
			if (err != CL_SUCCESS)
			{
	        		printf("Error: Failed to access device number of compute units !\n");
	        		return 1;
			}
			printf("with a max of %d compute units \n",comp_unit);

			//get local memory size on device
			cl_ulong tmpLong;
			if(clGetDeviceInfo(device_id[d], CL_DEVICE_LOCAL_MEM_SIZE, sizeof(tmpLong), &tmpLong, NULL) != CL_SUCCESS) {
				printf("Unable to get local memory size\n");
				return 1;
			}
			//printf("Device local mem size: %ld\n", tmpLong);
			//printf("Input size: %lu ",sizeof(type)*elements);
		
			//get max memory allocation on device
			if(clGetDeviceInfo(device_id[d],  CL_DEVICE_MAX_MEM_ALLOC_SIZE , sizeof(tmpLong), &tmpLong, NULL) != CL_SUCCESS) {
				printf("Unable to get max memory size\n");
				return 1;
			}
			//printf("Device memory allocation size: %ld\n", tmpLong);
			//printf("Buffer size allocated %ld\n", sizeof(type) * elements);

			//get max no. of work items in a work group on a single compute unit
			err = clGetDeviceInfo( device_id[d], CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(size_t), &work_group_size, NULL); 
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to get device Info (CL_DEVICE_MAX_WORK_GROUP_SIZE)!\n",err);
				return 1;
			}
			//printf("max work group size: %lu\n", work_group_size);

			//get Maximum number of work-items that can be specified in each dimension of the work-group to clEnqueueNDRangeKernel
			err = clGetDeviceInfo( device_id[d], CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(size_t)*3, work_item_size, NULL); 
			if (err != CL_SUCCESS)
			{
				printf("Error: Failed to get device Info (CL_DEVICE_MAX_WORK_ITEM_SIZES)!\n",err);
				return 1;
			}
			//for(int i=0;i<3;i++)
				//printf("max work item size: %lu ", work_item_size[i]);
			//printf("\n");


			//(3) context properties list - must be terminated with 0
			properties[0]= CL_CONTEXT_PLATFORM;
			properties[1]= (cl_context_properties) platform_id;
			properties[2]= 0;

			// create a context with the GPU device
			context = clCreateContext(properties,1,&device_id[d],NULL,NULL,&err);
			//printf("context, %x, err %d, \n", context, err);

			//(4) create command queue using the context and device
			command_queue = clCreateCommandQueue(context, device_id[d], CL_QUEUE_PROFILING_ENABLE, &err);
			//printf("command queue, %x, err %d, \n", command_queue, err);
	
			//(5) Get Source code
			char * ProgramSource = myLoadProgSource("project.cl", "", NULL);
			if(!ProgramSource)
			{
				printf("Failed to load program source!\n");
				system("pause");
				return 1;	
			}

			//(6) create a program from the kernel source code
			program = clCreateProgramWithSource(context,1,(const char **) &ProgramSource, NULL, &err);
			printf("program, %x, err %d, \n", program, err);

			//(7) compile the program
			if ((err = clBuildProgram(program, 1, &device_id[d], NULL, NULL, NULL)) != CL_SUCCESS)
			{
				printf("Error building program, err %d\n", err);
				return 1;
			}
	
			//(8) specify which kernel from the program to execute
			kernel = clCreateKernel(program, "project", &err);
			//printf("return kernel %lx, err %d\n", kernel, err);

			//(9) create buffer for the input/ouput (one for both)
			clInput = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR, sizeof(type) * elements, inputData, &err);
			//printf("return input %lx, err %d\n", clInput, err);
			clOutput = clCreateBuffer(context, CL_MEM_WRITE_ONLY | CL_MEM_USE_HOST_PTR, sizeof(type) * elements, results, &err);
			//printf("return output %lx, err %d\n", clInput, err);

			

			//global = elements;
			local = work_group_size;//256;  
			global = 7*local*comp_unit; //7 wavefronts
			if(global < elements/4){
				while( (elements / 4) % global != 0 )
					global += local;
			}
			else
				global = elements;
	
			//printf("global:%d ",global);

			err = clSetKernelArg(kernel, 0, sizeof(void*), (void*)&clInput);
					//printf("Set arg err %d\n", err);
					err = clSetKernelArg(kernel, 1, sizeof(void*), (void*)&clOutput);
					//printf("Set arg err %d\n", err);
					err = clSetKernelArg(kernel, 2, sizeof(int), &elements);
					//printf("Set arg err %d\n", err);
					
					
					cl_ulong map_time_start = 0, map_time_end = 0, unmap_time_start = 0, unmap_time_end = 0;
					double map_time=0,unmap_time=0,map_time_taken=0,unmap_time_taken=0;
					int nloops = NLOOPS;
					double total_time=0;
					map_time = 0;
					type *output_ptr;

					QueryPerformanceFrequency(&frequency);
					QueryPerformanceCounter(&start);
					while(nloops--)
					{

						cl_ulong time_start=0, time_end=0;
						double time_taken=0;

						// enqueue the kernel command for execution
						err = clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, &global, &local, 0, NULL, &kernel_event); //using local size = Null for time being
						////printf("launch kernel err %d\n", err);
						clFinish(command_queue);
	

						// copy the results from out of the buffer
						//err = clEnqueueReadBuffer(command_queue, clOutput, CL_TRUE, 0, sizeof(type) *elements, results, 0, NULL, NULL);
						//printf("reading buffer err %d\n", err);

						output_ptr = (type*) clEnqueueMapBuffer (command_queue, clOutput, CL_TRUE, CL_MAP_READ, 0, sizeof(type) * elements, 0, NULL, &map_event, &err);
						//memcpy(results, output_ptr, sizeof(type)*elements);
						//clEnqueueUnmapMemObject (command_queue, clOutput , output_ptr, 0,  NULL, &unmap_event); 

						clGetEventProfilingInfo(kernel_event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start, NULL);
						clGetEventProfilingInfo(kernel_event, CL_PROFILING_COMMAND_END, sizeof(time_end), &time_end, NULL);
						time_taken = time_end - time_start;
						total_time += time_taken;

						clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_START, sizeof(map_time_start), &map_time_start, NULL);
						clGetEventProfilingInfo(map_event, CL_PROFILING_COMMAND_END, sizeof(map_time_end), &map_time_end, NULL);
						map_time_taken = map_time_end - map_time_start;
						map_time +=map_time_taken;
					}
  					QueryPerformanceCounter(&stop);

					printf("map event time:%f ns\n",map_time/NLOOPS);

					double timespent = (double)(stop.QuadPart - start.QuadPart) / (double)frequency.QuadPart;
					printf ("Project on GPU took total = %f seconds\n", timespent/NLOOPS);

			

					total_time = total_time/NLOOPS;
					printf("GPU projection took total time (event profiling)= %f seconds\n", (total_time / 1e9) );

					// print the results
					int c = 0;
					for(int i=0; i<elements; i++)
					{
						int a = output_ptr[i] - result_cpu[i];
						if(a!=0)
							c++;
						//printf("input:%u ",inputData[i]);
						//printf("output: %u ", results[i]);
	
					}
					printf("Difference count:%d ",c);
					printf("\n");
					//clEnqueueUnmapMemObject (command_queue, clOutput, output_ptr, 0,  NULL, &unmap_event); 
					//double sec = 
					//sec = sec/NLOOPS;
					//printf ("GPU  projection took total time of day= %f seconds\n", sec); //(for timeofday)
			

					size_t elementsAccessed = 2 * elements;
					float GBs = (((elementsAccessed) * sizeof(type))/ (total_time / 1e9)) / ((float)(1<<30));
					std::cout << " " << GBs << " GBs\n";

	
					delete [] inputData;
					delete [] results;
			}
		}
	
			// cleanup - release OpenCL resources
			clReleaseMemObject(clInput);
			clReleaseMemObject(clOutput);
			clReleaseProgram(program);
			clReleaseKernel(kernel);
			clReleaseCommandQueue(command_queue);
			clReleaseContext(context);

			system("pause");
			return 0;
}



